#
# Copyright (c) 2015-2023 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_*** module

"""

from pyramid.interfaces import IRequest

from pyams_app_msc.shared.catalog import ICatalogEntryInfo, IWfCatalogEntry
from pyams_content_api.feature.json import IJSONExporter, JSONBaseExporter
from pyams_utils.adapter import adapter_config

__docformat__ = 'restructuredtext'


@adapter_config(name='catalog_entry',
                required=(IWfCatalogEntry, IRequest),
                provides=IJSONExporter)
class JSONCatalogEntryExporter(JSONBaseExporter):
    """Catalog entry API info"""
    
    is_inner = True
    conversion_target = None

    def convert_content(self, **params):
        """JSON catalog entry conversion"""
        result = super().convert_content(**params)
        info = ICatalogEntryInfo(self.context)
        for name in self.context.field_names:
            self.get_attribute(result, name, context=info)
        return result



#
# Copyright (c) 2015-2023 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_*** module

"""

from datetime import datetime, timedelta, timezone

from persistent import Persistent
from pyramid.authorization import ALL_PERMISSIONS, Allow
from pyramid.events import subscriber
from pyramid.interfaces import IRequest, IView
from pyramid.renderers import render
from pyramid.response import Response
from zope.container.contained import Contained
from zope.interface import implementer
from zope.lifecycleevent import ObjectModifiedEvent
from zope.lifecycleevent.interfaces import IObjectModifiedEvent
from zope.schema.fieldproperty import FieldProperty

from pyams_app_msc.feature.booking.interfaces import BOOKING_STATUS, IBookingContainer, IBookingInfo, \
    IBookingReminderTask, IBookingTarget
from pyams_app_msc.feature.booking.message import get_booking_message_values
from pyams_app_msc.feature.messaging.interfaces import IMessagingSettings
from pyams_app_msc.feature.planning.interfaces import IPlanning, ISession
from pyams_app_msc.feature.profile.interfaces import IUserProfile
from pyams_app_msc.feature.quotation import Quotation
from pyams_app_msc.interfaces import CANCEL_BOOKING_PERMISSION, MANAGE_BOOKING_PERMISSION
from pyams_app_msc.shared.catalog import ICatalogEntry
from pyams_app_msc.shared.theater.interfaces import BOOKING_CANCEL_MODE, IMovieTheater, IMovieTheaterSettings
from pyams_app_msc.shared.theater.interfaces.audience import ICinemaAudienceContainer
from pyams_app_msc.shared.theater.interfaces.mail import IMailTemplates
from pyams_app_msc.shared.theater.interfaces.price import ICinemaPriceContainer
from pyams_catalog.utils import index_object
from pyams_content.feature.history import IHistoryContainer
from pyams_content.feature.history.interfaces import IHistoryTarget
from pyams_content.interfaces import IObjectType
from pyams_content_api.feature.json import IJSONExporter, JSONBaseExporter
from pyams_file.property import FileProperty
from pyams_i18n.interfaces import II18n
from pyams_mail.interfaces import IPrincipalMailInfo
from pyams_mail.message import HTMLMessage
from pyams_scheduler.interfaces import IScheduler
from pyams_scheduler.interfaces.task import IDateTaskScheduling, SCHEDULER_TASK_DATE_MODE
from pyams_security.interfaces import ISecurityManager, IViewContextPermissionChecker
from pyams_security.interfaces.base import FORBIDDEN_PERMISSION
from pyams_security.interfaces.names import ADMIN_USER_ID
from pyams_security.interfaces.notification import INotificationSettings
from pyams_security.security import ProtectedObjectMixin
from pyams_security.utility import get_principal
from pyams_site.interfaces import ISiteRoot
from pyams_utils.adapter import ContextAdapter, adapter_config
from pyams_utils.factory import create_object, factory_config, get_interface_base_name
from pyams_utils.interfaces import ICacheKeyValue
from pyams_utils.registry import get_pyramid_registry, get_utility, query_utility
from pyams_utils.request import check_request
from pyams_utils.timezone import tztime
from pyams_utils.traversing import get_parent
from pyams_utils.zodb import load_object, volatile_property
from pyams_workflow.interfaces import IWorkflowVersions
from pyams_zmi.utils import get_object_label

__docformat__ = 'restructuredtext'

from pyams_app_msc import _


@implementer(IView)
class FakeView:
    """Fake view"""

    def __call__(self, context, request):
        return Response('')

    
@factory_config(IBookingInfo)
@implementer(IHistoryTarget)
class BookingInfo(ProtectedObjectMixin, Persistent, Contained):
    """Booking information"""

    creator = FieldProperty(IBookingInfo['creator'])
    recipient = FieldProperty(IBookingInfo['recipient'])
    recipient_establishment = FieldProperty(IBookingInfo['recipient_establishment'])
    status = FieldProperty(IBookingInfo['status'])
    nb_participants = FieldProperty(IBookingInfo['nb_participants'])
    participants_age = FieldProperty(IBookingInfo['participants_age'])
    nb_accompanists = FieldProperty(IBookingInfo['nb_accompanists'])
    nb_free_accompanists = FieldProperty(IBookingInfo['nb_free_accompanists'])
    accompanying_ratio = FieldProperty(IBookingInfo['accompanying_ratio'])
    nb_groups = FieldProperty(IBookingInfo['nb_groups'])
    price = FieldProperty(IBookingInfo['price'])
    cultural_pass = FieldProperty(IBookingInfo['cultural_pass'])
    comments = FieldProperty(IBookingInfo['comments'])
    notepad = FieldProperty(IBookingInfo['notepad'])
    archived = FieldProperty(IBookingInfo['archived'])
    quotation_number = FieldProperty(IBookingInfo['quotation_number'])
    quotation_message = FieldProperty(IBookingInfo['quotation_message'])
    quotation = FileProperty(IBookingInfo['quotation'])
    send_update = FieldProperty(IBookingInfo['send_update'])
    update_subject = FieldProperty(IBookingInfo['update_subject'])
    update_message = FieldProperty(IBookingInfo['update_message'])
    send_reminder = FieldProperty(IBookingInfo['send_reminder'])
    reminder_subject = FieldProperty(IBookingInfo['reminder_subject'])
    reminder_message = FieldProperty(IBookingInfo['reminder_message'])
    reminder_date = FieldProperty(IBookingInfo['reminder_date'])

    def __acl__(self):
        result = [
            (Allow, ADMIN_USER_ID, ALL_PERMISSIONS)
        ]
        theater = get_parent(self, IMovieTheater)
        settings = IMovieTheaterSettings(theater)
        if settings.booking_cancel_mode != BOOKING_CANCEL_MODE.FORBIDDEN.value:
            result.append((Allow, self.recipient, CANCEL_BOOKING_PERMISSION))
        return result

    def get_recipient(self, with_info=True):
        """Recipient principal getter"""
        principal = get_principal(None, principal_id=self.recipient)
        if with_info:
            profile = IUserProfile(principal, None)
            return principal, profile
        return principal

    @property
    def nb_seats(self):
        """Total seats count"""
        return (self.nb_participants or 0) + (self.nb_accompanists or 0)

    def get_price(self):
        """Price getter"""
        if not self.price:
            return None
        theater = IMovieTheater(self)
        return ICinemaPriceContainer(theater).get(self.price)

    @volatile_property
    def session(self):
        """Booking session getter"""
        return get_parent(self, IBookingTarget)

    @property
    def session_index(self):
        """Booking session index value getter"""
        return ICacheKeyValue(self.session)

    def set_session(self, new_session):
        """Update booking session"""
        if not ISession.providedBy(new_session):
            new_session = load_object(new_session, self)
        if (new_session is self.session) or not ISession.providedBy(new_session):
            return
        old_bookings = self.__parent__
        new_bookings = IBookingContainer(new_session)
        new_bookings.append(self)
        index_object(self)
        del self.session
        if self.quotation is not None:
            del self.quotation
            self.get_quotation(force_refresh=True)
        del old_bookings[ICacheKeyValue(self)]
        registry = get_pyramid_registry()
        registry.notify(ObjectModifiedEvent(self))
        return self

    def get_quotation_number(self):
        theater = IMovieTheater(self)
        settings = IMovieTheaterSettings(theater)
        return settings.get_quotation_number()

    def get_quotation(self, force_refresh=False, store=True, **kwargs):
        quotation = self.quotation
        if force_refresh or (not quotation):
            quotation_number = self.quotation_number
            if not quotation_number:
                quotation_number = self.quotation_number = self.get_quotation_number()
            quotation = (f'{quotation_number}.pdf',
                         bytes(Quotation(self, quotation_number, **kwargs)))
            if store:
                self.quotation = quotation
        return quotation

    def send_request_notification_message(self, audience, view):
        """Send request confirmation message to recipient and to theater operator"""
        request = check_request()
        settings = IMessagingSettings(request.root, None)
        if settings is None:
            return None
        mailer = settings.get_mailer()
        if mailer is None:
            return None
        translate = request.localizer.translate
        # get target email
        sm = get_utility(ISecurityManager)
        target_email = None
        profile = IUserProfile(request.principal, None)
        if profile is not None:
            target_email = profile.email
        if not target_email:
            principal = sm.get_raw_principal(request.principal.id)
            mail_info = IPrincipalMailInfo(principal, None)
            if mail_info is not None:
                target_email = [
                    f'{name} <{address}>'
                    for name, address in mail_info.get_addresses()
                ]
                if len(target_email) > 0:
                    target_email = target_email[0]
        # get contact email
        theater = IMovieTheater(self)
        contact_email = None
        audience = ICinemaAudienceContainer(theater).get(audience)
        if audience is not None:
            contact = audience.contact
            if contact is not None:
                contact_email = f'{contact.name} <{contact.email_address}>' if contact.email_address else None
        if not contact_email:
            contact_email = theater.contact_email
        notifications_settings = INotificationSettings(sm)
        # send notification message
        if contact_email:
            message_body = render('templates/booking-new-notify.pt',
                                  request=request,
                                  value={
                                      'settings': notifications_settings,
                                      'theater': theater,
                                      'booking': self,
                                      'session': self.session,
                                      'title': get_object_label(self.session, request, view,
                                                                name='short-text'),
                                      'sender': sm.get_principal(request.principal.id)
                                  })
            message = HTMLMessage(subject=translate(_("{} Booking request notification"))
                                      .format(settings.subject_prefix),
                                  from_addr=f'{settings.source_name} <{settings.source_address}>',
                                  to_addr=contact_email,
                                  reply_to=target_email,
                                  html=message_body)
            mailer.send(message)
        # send confirmation message
        if getattr(self, 'send_confirmation', False):
            if target_email:
                # create inner message
                message_body = render('templates/booking-new-ack.pt',
                                      request=request,
                                      value={
                                          'settings': notifications_settings,
                                          'booking': self,
                                          'session': self.session,
                                          'title': get_object_label(self.session, request, view,
                                                                    name='short-text'),
                                          'sender': sm.get_principal(request.principal.id)
                                      })
                message = HTMLMessage(subject=translate(_("{} Booking request confirmation"))
                                          .format(settings.subject_prefix),
                                      from_addr=f'{settings.source_name} <{settings.source_address}>',
                                      to_addr=target_email,
                                      reply_to=contact_email,
                                      html=message_body)
                mailer.send(message)

    def send_reminder_message(self):
        """Send reminder message"""
        if not (self.send_reminder and self.reminder_message):
            return
        root = get_parent(self, ISiteRoot)
        settings = IMessagingSettings(root, None)
        if settings is None:
            return
        mailer = settings.get_mailer()
        if mailer is None:
            return
        sm = get_utility(ISecurityManager)
        principal = sm.get_raw_principal(self.recipient)
        mail_info = IPrincipalMailInfo(principal, None)
        if mail_info is None:
            return
        mail_addresses = [
            f'{name} <{address}>'
            for name, address in mail_info.get_addresses()
        ]
        message = HTMLMessage(f'{settings.subject_prefix} {self.reminder_subject}',
                              from_addr=f'{settings.source_name} <{settings.source_address}>',
                              to_addr=mail_addresses,
                              html=self.reminder_message)
        mailer.send(message)
        self.reminder_date = tztime(datetime.now(timezone.utc))


@adapter_config(required=IBookingInfo,
                provides=IObjectType)
def booking_object_type(context):
    """Booking object type value adapter"""
    return get_interface_base_name(IBookingInfo)


@adapter_config(required=IBookingInfo,
                provides=IViewContextPermissionChecker)
class BookingInfoPermissionChecker(ContextAdapter):
    """Booking permission checker"""

    @property
    def edit_permission(self):
        if self.context.archived:
            return FORBIDDEN_PERMISSION
        return MANAGE_BOOKING_PERMISSION


@adapter_config(name='delete',
                required=IBookingInfo,
                provides=IViewContextPermissionChecker)
class BookingInfoDeletePermissionChecker(ContextAdapter):
    """Booking delete permission checker"""

    @property
    def edit_permission(self):
        if self.context.archived or (self.context.status == BOOKING_STATUS.ACCEPTED.value):
            return FORBIDDEN_PERMISSION
        return MANAGE_BOOKING_PERMISSION


@adapter_config(required=IBookingInfo,
                provides=IBookingContainer)
def booking_container_adapter(context):
    """Booking container adapter"""
    return context.__parent__


@adapter_config(required=IBookingInfo,
                provides=ISession)
def booking_session_adapter(context):
    """Booking session adapter"""
    return context.session


@adapter_config(required=IBookingInfo,
                provides=IPlanning)
def booking_planning_adapter(context):
    """Booking planning adapter"""
    return get_parent(context, IPlanning)


@adapter_config(required=IBookingInfo,
                provides=IMovieTheater)
def booking_theater_adapter(context):
    """Booking theater adapter"""
    return get_parent(context, IMovieTheater)


@adapter_config(required=IBookingInfo,
                provides=IUserProfile)
def booking_user_profile_adapter(context):
    """Booking user profile adapter"""
    _, profile = context.get_recipient()
    return profile


@subscriber(IObjectModifiedEvent, context_selector=IBookingInfo)
def handle_modified_booking(event):
    """Handle modified booking"""
    scheduler = query_utility(IScheduler)
    if scheduler is None:
        return
    booking = IBookingInfo(event.object)
    theater = IMovieTheater(booking)
    task_id = f'reminder::{ICacheKeyValue(booking)}'
    if task_id in scheduler:
        del scheduler[task_id]
    settings = IMovieTheaterSettings(theater)
    if not settings.reminder_delay:
        return
    reminder_date = booking.session.start_date - timedelta(days=settings.reminder_delay)
    if reminder_date < tztime(datetime.now(timezone.utc)):
        return
    if booking.send_reminder and not booking.reminder_date:
        task = create_object(IBookingReminderTask, booking=booking)
        task.name = f'Booking {task_id}'
        task.schedule_mode = SCHEDULER_TASK_DATE_MODE
        scheduler_info = IDateTaskScheduling(task)
        scheduler_info.start_date = reminder_date
        scheduler_info.active = True
        scheduler[task_id] = task


@adapter_config(name='msc_mobile_api',
                required=(IBookingInfo, IRequest),
                provides=IJSONExporter)
class JSONBookingInfoSessionExporter(JSONBaseExporter):
    """JSON booking info session exporter"""
    
    conversion_target = None
    
    def convert_content(self, **params):
        result = super().convert_content(**params)
        context = self.context
        result['id'] = ICacheKeyValue(context)
        recipient = get_principal(principal_id=context.recipient)
        session = ISession(context)
        if params.get('with_session_info'):
            request = self.request
            registry = request.registry
            exporter = registry.queryMultiAdapter((session, request), IJSONExporter,
                                                  name='msc_mobile_api')
            if exporter is not None:
                result['session'] = exporter.to_json()
        entry = get_parent(session, ICatalogEntry)
        if entry is not None:
            version = IWorkflowVersions(entry).get_last_versions()[0]
            self.get_attribute(result,
                               name='activity',
                               context=version,
                               getter=lambda x, _attr: II18n(x).query_attribute('title'))
        else:
            self.get_attribute(result,
                               name='activity',
                               context=session,
                               getter=lambda x, _attr: x.label)
        theater = get_parent(session, IMovieTheater)
        self.get_attribute(result,
                           name='theater',
                           context=theater,
                           getter=lambda x, _attr: II18n(x).query_attribute('title'))
        self.get_attribute(result, 'recipient',
                           context=recipient,
                           getter=lambda x, _attr: x.title)
        profile = IUserProfile(recipient, None)
        if profile is not None:
            self.get_attribute(result,
                               name='establishment',
                               context=profile,
                               getter=lambda x, _attr: x.establishment)
        self.get_attribute(result, 'status')
        self.get_attribute(result, 'nb_participants')
        self.get_attribute(result, 'participants_age')
        self.get_attribute(result, 'nb_accompanists')
        self.get_attribute(result, 'accompanying_ratio')
        self.get_attribute(result, 'nb_groups')
        self.get_attribute(result, 'cultural_pass')
        self.get_attribute(result, 'price')
        self.get_attribute(result,
                           name='price_label',
                           context=context.get_price(),
                           getter=lambda x, _attr: II18n(x).query_attribute('name') if x is not None else None)
        self.get_attribute(result, 'comments')
        history = IHistoryContainer(context, None)
        if history is not None:
            result['notepads'] = [
                comment.comment
                for comment in history.values()
                if comment.comment
            ]
        self.get_attribute(result,
                           name='can_update',
                           getter=lambda x, _attr: self.request.has_permission(MANAGE_BOOKING_PERMISSION,
                                                                               context=x))
        return result


@adapter_config(name='msc_mobile_validation_api',
                required=(IBookingInfo, IRequest),
                provides=IJSONExporter)
class JSONBookingInfoValidationExporter(JSONBookingInfoSessionExporter):
    """JSON booking info validation exporter"""
    
    conversion_target = None

    def get_prices(self, _context, _attr):
        """Available prices getter"""
        theater = IMovieTheater(self.context)
        return [
            {
                'value': price.__name__,
                'label': price.name
            }
            for price in ICinemaPriceContainer(theater).get_active_items()
        ]

    def convert_content(self, **params):
        result = super().convert_content(**params)
        context = self.context
        self.get_attribute(result,
                           name='available_prices',
                           getter=self.get_prices)
        values = get_booking_message_values(context, self.request, FakeView())
        theater = IMovieTheater(context)
        templates = IMailTemplates(theater)
        template = templates.accept_subject
        if template:
            result['notify_subject'] = template.format(**values)
        template = templates.accept_template
        if template:
            result['notify_message'] = template.format(**values)
        settings = IMovieTheaterSettings(theater)
        if settings.reminder_delay:
            session = ISession(context)
            reminder_date = session.start_date - timedelta(days=settings.reminder_delay)
            can_send_reminder = result['can_send_reminder'] = reminder_date >= tztime(datetime.now(timezone.utc))
            if can_send_reminder:
                template = templates.reminder_subject
                if template:
                    result['reminder_subject'] = template.format(**values)
                template = templates.reminder_template
                if template:
                    result['reminder_message'] = template.format(**values)
        else:
            result['can_send_reminder'] = False
        return result
    


#
# Copyright (c) 2015-2023 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_*** module

"""

import json
from datetime import datetime, timezone

from hypatia.catalog import CatalogQuery
from hypatia.interfaces import ICatalog
from hypatia.query import And, Eq, Ge, Le
from pyramid.decorator import reify
from pyramid.httpexceptions import HTTPBadRequest, HTTPNotFound
from pyramid.view import view_config
from zope.interface import implementer

from pyams_app_msc.feature.booking import IBookingContainer
from pyams_app_msc.feature.planning import IPlanning, IPlanningTarget, IWfPlanningTarget
from pyams_app_msc.feature.planning.interfaces import ISession
from pyams_app_msc.feature.planning.zmi.interfaces import IActivityArchivedSessionsTable, IActivityCurrentSessionsTable, \
    IActivitySessionsTable, IPlanningMenu
from pyams_app_msc.interfaces import MANAGE_PLANNING_PERMISSION, VIEW_BOOKING_PERMISSION, \
    VIEW_PLANNING_PERMISSION
from pyams_app_msc.shared.theater import IMovieTheaterSettings
from pyams_app_msc.shared.theater.interfaces import IMovieTheater
from pyams_app_msc.shared.theater.interfaces.room import ICinemaRoomContainer
from pyams_catalog.query import CatalogResultSet
from pyams_content_api.feature.json.interfaces import IJSONExporter
from pyams_layer.interfaces import IPyAMSLayer
from pyams_pagelet.pagelet import pagelet_config
from pyams_security.interfaces.base import USE_INTERNAL_API_PERMISSION
from pyams_skin.viewlet.actions import ContextAction, JsContextAction
from pyams_table.column import GetAttrColumn, GetAttrFormatterColumn
from pyams_table.interfaces import IColumn, IValues
from pyams_template.template import template_config
from pyams_utils.adapter import ContextRequestViewAdapter, adapter_config
from pyams_utils.factory import factory_config, get_interface_base_name
from pyams_utils.interfaces import MISSING_INFO
from pyams_utils.interfaces.data import IObjectData
from pyams_utils.registry import get_utility
from pyams_utils.timezone import tztime
from pyams_utils.traversing import get_parent
from pyams_utils.url import absolute_url
from pyams_viewlet.manager import viewletmanager_config
from pyams_viewlet.viewlet import ViewContentProvider, viewlet_config
from pyams_zmi.interfaces import IAdminLayer
from pyams_zmi.interfaces.table import IColumnSortData, IInnerTable, ITableElementEditor
from pyams_zmi.interfaces.viewlet import IContentManagementMenu, ISecondaryActionsViewletManager, IViewWithoutToolbar
from pyams_zmi.table import I18nColumnMixin, IconColumn, InnerTableAdminView, MultipleTablesAdminView, \
    Table, TableElementEditor, get_table_id
from pyams_zmi.view import InnerAdminView
from pyams_zmi.zmi.viewlet.menu import NavigationMenuItem

__docformat__ = 'restructuredtext'

from pyams_app_msc import _


@viewletmanager_config(name='planning.menu',
                       context=IWfPlanningTarget, layer=IAdminLayer,
                       manager=IContentManagementMenu, weight=20,
                       provides=IPlanningMenu,
                       permission=VIEW_PLANNING_PERMISSION)
class PlanningMenu(NavigationMenuItem):
    """Planning menu"""

    label = _("Planning")
    icon_class = 'fas fa-calendar-week'
    href = '#planning.html'


@pagelet_config(name='planning.html',
                context=IWfPlanningTarget, layer=IPyAMSLayer,
                permission=VIEW_PLANNING_PERMISSION)
@template_config(template='templates/planning.pt',
                 layer=IPyAMSLayer)
class PlanningView(InnerAdminView):
    """Planning view"""

    title = _("Sessions planning")

    @reify
    def theater(self):
        """Theater getter"""
        return get_parent(self.context, IMovieTheater)

    @property
    def rooms(self):
        """Theater rooms iterator getter"""
        theater = self.theater
        if theater is not None:
            yield from ICinemaRoomContainer(theater).get_active_items()

    def get_context(self):
        """View context getter"""
        return IPlanning(self.context)

    @property
    def can_view_bookings(self):
        """Bookings access checker"""
        return bool(self.request.has_permission(VIEW_BOOKING_PERMISSION, context=self.context))

    @property
    def can_edit_planning(self):
        """Calendar editor checker"""
        return bool(self.request.has_permission(MANAGE_PLANNING_PERMISSION, context=self.context))

    def get_calendar_options(self, room):
        """Calendar options getter"""
        context = self.get_context()
        request = self.request
        settings = IMovieTheaterSettings(self.theater)
        can_edit = request.has_permission(MANAGE_PLANNING_PERMISSION, context=context)
        translate = request.localizer.translate
        options = {
            'editable': can_edit,
            'droppable': can_edit,
            'eventSources': [{
                'url': absolute_url(context, request, 'get-planning-events.json'),
                'extraParams': {
                    'room': room.__name__
                }
            }],
            'height': '100%',
            'firstDay': settings.calendar_first_day,
            'dateClick': 'MyAMS.msc.calendar.addEvent',
            'initialView': 'timeGridWeek',
            'headerToolbar': {
                'center': 'today',
                'right': 'prev,next dayGridMonth,timeGridWeek,timeGridDay,listMonth'
            },
            'allDaySlot': False,
            'slotDuration': f'00:{settings.calendar_slot_duration:#02}:00',
            'slotMinTime': room.start_time.isoformat(),
            'slotMaxTime': room.end_time.isoformat(),
            'buttonText': {
                'today': translate(_("Today")),
                'month': translate(_("Month")),
                'week': translate(_("Week")),
                'day': translate(_("Day")),
                'list': translate(_("List")),
                'all-day': translate(_("All-day")),
                'prev': " « ",
                'next': " » "
            }
        }
        return json.dumps(options)


@view_config(name='get-planning-events.json',
             context=IPlanning, request_type=IPyAMSLayer,
             permission=USE_INTERNAL_API_PERMISSION,
             renderer='json')
def get_planning_events(request):
    """Planning events getter"""
    params = request.params
    room = params.get('room')
    start = params.get('start')
    end = params.get('end')
    if not (room and start and end):
        raise HTTPBadRequest()
    theater = get_parent(request.context, IMovieTheater)
    if theater is None:
        raise HTTPNotFound()
    container = ICinemaRoomContainer(theater, None)
    if (container is None) or (room not in container):
        raise HTTPNotFound()
    catalog = get_utility(ICatalog)
    query = And(Eq(catalog['object_types'], get_interface_base_name(ISession)),
                Eq(catalog['planning_room'], room),
                Le(catalog['planning_start_date'], datetime.fromisoformat(end)),
                Ge(catalog['planning_end_date'], datetime.fromisoformat(start)))
    events = []
    for session in CatalogResultSet(CatalogQuery(catalog).query(query)):
        exporter = request.registry.queryMultiAdapter((session, request), IJSONExporter)
        if exporter is not None:
            events.append(exporter.to_json(with_edit_info=True, edit_context=request.context))
    return events


@viewlet_config(name='planning.transpose',
                layer=IAdminLayer, view=PlanningView,
                manager=ISecondaryActionsViewletManager, weight=10)
class PlanningTransposeAction(JsContextAction):
    """Planning transpose action"""

    icon_class = 'fas fa-recycle'
    hint = _("Transpose calendars")
    hint_placement = 'bottom'

    href = 'MyAMS.msc.calendar.transpose'


@viewlet_config(name='planning.synchronize',
                layer=IAdminLayer, view=PlanningView,
                manager=ISecondaryActionsViewletManager, weight=20)
class PlanningSynchronizeAction(JsContextAction):
    """Planning synchronize action"""

    icon_class = 'fas fa-sync-alt'
    hint = _("Synchronize calendars views")
    hint_placement = 'bottom'

    href = 'MyAMS.msc.calendar.synchronize'


@viewlet_config(name='planning.scroll',
                layer=IAdminLayer, view=PlanningView,
                manager=ISecondaryActionsViewletManager, weight=30)
class PlanningScrollAction(JsContextAction):
    """Planning scrolling action"""

    icon_class = 'fas fa-arrows-alt-v'
    hint = _("Synchronize calendars scrolling")
    hint_placement = 'bottom'

    href = 'MyAMS.msc.calendar.scroll'


@viewlet_config(name='planning-legend.action',
                layer=IAdminLayer, view=PlanningView,
                manager=ISecondaryActionsViewletManager, weight=40)
@template_config(template='templates/planning-legend.pt',
                 layer=IAdminLayer)
class PlanningLegendAction(ContextAction):
    """Planning legend action"""
    
    icon_class = 'fas fa-question-circle text-info'
    hint = _("Planning legend")
    hint_placement = 'top'


#
# Sessions view
#

@viewlet_config(name='activity-sessions.menu',
                context=IWfPlanningTarget, layer=IAdminLayer,
                manager=IPlanningMenu, weight=10,
                permission=VIEW_PLANNING_PERMISSION)
class ActivitySessionsMenu(NavigationMenuItem):
    """Activity sessions menu"""

    label = _("Planned sessions")
    href = '#activity-sessions.html'


@pagelet_config(name='activity-sessions.html',
                context=IWfPlanningTarget, layer=IPyAMSLayer,
                permission=VIEW_PLANNING_PERMISSION)
class ActivitySessionsView(MultipleTablesAdminView):
    """Activity sessions view"""

    table_label = _("Activity planned sessions")


class ActivitySessionsBaseTableView(InnerTableAdminView, ViewContentProvider):
    """Activity base sessions view"""
    
    hide_section = True

    
@implementer(IViewWithoutToolbar, IObjectData)
class ActivitySessionsBaseTable(Table):
    """Activity base sessions table"""

    sort_index = 1

    @reify
    def id(self):  # pylint: disable=invalid-name
        """Table ID getter"""
        planning = IPlanning(self.context)
        return get_table_id(self, planning)

    @reify
    def data_attributes(self):
        attributes = super().data_attributes
        attributes.setdefault('table', {}).update({
            'data-ams-order': f'{self.sort_index},{self.sort_order}'
        })
        return attributes

    object_data = {
        'buttons': ['copy', 'csv', 'excel', 'print'],
        'ams-buttons-classname': 'btn btn-sm btn-secondary'
    }

    @property
    def css_classes(self):
        classes = super().css_classes.copy()
        classes['tr.selected'] = self.get_tr_class
        return classes

    def get_tr_class(self, item, column, index):
        now = tztime(datetime.now(timezone.utc))
        if item.start_date < now:
            return 'bg-warning'
        return None


@adapter_config(name='current-sessions',
                required=(IWfPlanningTarget, IAdminLayer, ActivitySessionsView),
                provides=IInnerTable)
class ActivityCurrentSessionsTableView(ActivitySessionsBaseTableView):
    """Activity current sessions view"""

    weight = 10

    table_class = IActivityCurrentSessionsTable
    table_label = _("Incoming sessions")


@factory_config(IActivityCurrentSessionsTable)
class ActivityCurrentSessionsTable(ActivitySessionsBaseTable):
    """Activity current sessions table"""

    prefix = 'current_sessions_table'
    sort_order = 'asc'


@adapter_config(required=(IPlanningTarget, IAdminLayer, IActivityCurrentSessionsTable),
                provides=IValues)
@adapter_config(required=(IWfPlanningTarget, IAdminLayer, IActivityCurrentSessionsTable),
                provides=IValues)
class ActivityCurrentSessionsTableValues(ContextRequestViewAdapter):
    """Activity current sessions table values adapter"""

    @property
    def values(self):
        now = tztime(datetime.now(timezone.utc))
        yield from (
            session
            for session in IPlanning(self.context).values()
            if session.start_date >= now
        )


@adapter_config(name='archived-sessions',
                required=(IWfPlanningTarget, IAdminLayer, ActivitySessionsView),
                provides=IInnerTable)
class ActivityArchivedSessionsTableView(ActivitySessionsBaseTableView):
    """Activity archived sessions view"""

    weight = 20

    table_class = IActivityArchivedSessionsTable
    table_label = _("Archived sessions")


@factory_config(IActivityArchivedSessionsTable)
class ActivityArchivedSessionsTable(ActivitySessionsBaseTable):
    """Activity archived sessions table"""

    prefix = 'archived_sessions_table'
    sort_order = 'desc'

    
@adapter_config(required=(IPlanningTarget, IAdminLayer, IActivityArchivedSessionsTable),
                provides=IValues)
@adapter_config(required=(IWfPlanningTarget, IAdminLayer, IActivityArchivedSessionsTable),
                provides=IValues)
class ActivityArchivedSessionsTableValues(ContextRequestViewAdapter):
    """Activity archived sessions table values adapter"""

    @property
    def values(self):
        now = tztime(datetime.now(timezone.utc))
        yield from (
            session
            for session in IPlanning(self.context).values()
            if session.start_date < now
        )


@adapter_config(name='archived',
                required=(IWfPlanningTarget, IAdminLayer, IActivitySessionsTable),
                provides=IColumn)
class ActivitySessionsTableArchivedColumn(IconColumn):
    """Activity sessions table archived column"""
    
    css_classes = {
        'th': 'action',
        'td': 'text-danger'
    }
    icon_class = 'fas fa-archive'
    hint = _("Archived")

    weight = 5

    def checker(self, item):
        now = tztime(datetime.now(timezone.utc))
        return item.start_date < now


@adapter_config(name='start_date',
                required=(IWfPlanningTarget, IAdminLayer, IActivitySessionsTable),
                provides=IColumn)
@implementer(IColumnSortData)
class ActivitySessionsTableStartDateColumn(I18nColumnMixin, GetAttrFormatterColumn):
    """Activity sessions table start date column"""

    i18n_header = _("Start date")
    attr_name = 'start_date'
    
    @property
    def format_string(self):
        translate = self.request.localizer.translate
        return translate(_("%a %d/%m/%Y at %H:%M"))

    weight = 10
    
    @staticmethod
    def get_sort_value(value):
        return value.start_date.isoformat()


@adapter_config(name='room',
                required=(IWfPlanningTarget, IAdminLayer, IActivitySessionsTable),
                provides=IColumn)
class ActivitySessionsTableRoomColumn(I18nColumnMixin, GetAttrColumn):
    """Activity sessions table room column"""

    i18n_header = _("Room")
    attr_name = 'room'

    weight = 20

    def get_value(self, obj):
        room = obj.get_room()
        return room.name if room is not None else MISSING_INFO


@adapter_config(name='capacity',
                required=(IWfPlanningTarget, IAdminLayer, IActivitySessionsTable),
                provides=IColumn)
class ActivitySessionsTableCapacityColumn(I18nColumnMixin, GetAttrColumn):
    """Activity sessions table capacity column"""

    i18n_header = _("Session capacity")
    attr_name = 'capacity'

    weight = 30


@adapter_config(name='bookings',
                required=(IWfPlanningTarget, IAdminLayer, IActivitySessionsTable),
                provides=IColumn)
class ActivitySessionsTableBookingsColumn(I18nColumnMixin, GetAttrColumn):
    """Activity sessions table bookings column"""

    i18n_header = _("Reserved")
    attr_name = 'bookings'

    weight = 40

    def get_value(self, obj):
        return IBookingContainer(obj).get_requested_seats()


@adapter_config(name='free_seats',
                required=(IWfPlanningTarget, IAdminLayer, IActivityCurrentSessionsTable),
                provides=IColumn)
class ActivitySessionsTableFreeSeatsColumn(I18nColumnMixin, GetAttrColumn):
    """Activity sessions table free seats column"""

    i18n_header = _("Free seats")
    attr_name = 'free_seats'

    weight = 50

    def get_value(self, obj):
        return IBookingContainer(obj).free_seats


@adapter_config(name='confirmed',
                required=(IWfPlanningTarget, IAdminLayer, IActivityArchivedSessionsTable),
                provides=IColumn)
class ActivitySessionsTableConfirmedColumn(I18nColumnMixin, GetAttrColumn):
    """Activity sessions table confirmed column"""

    i18n_header = _("Confirmed")
    attr_name = 'confirmed'

    weight = 50

    def get_value(self, obj):
        return IBookingContainer(obj).get_confirmed_seats()


@adapter_config(name='temporary',
                required=(IWfPlanningTarget, IAdminLayer, IActivitySessionsTable),
                provides=IColumn)
class ActivitySessionsTableWaitingColumn(IconColumn):
    """Activity sessions table waiting column"""

    icon_class = 'fas fa-hourglass'
    hint = _("Temporary session")

    weight = 60

    def checker(self, item):
        return item.temporary


@adapter_config(name='bookable',
                required=(IWfPlanningTarget, IAdminLayer, IActivitySessionsTable),
                provides=IColumn)
class ActivitySessionsTableBookableColumn(IconColumn):
    """Activity sessions table bookable column"""

    icon_class = 'far fa-calendar'
    hint = _("Open for internal booking")

    weight = 70

    def checker(self, item):
        return item.bookable


@adapter_config(name='extern-bookable',
                required=(IWfPlanningTarget, IAdminLayer, IActivitySessionsTable),
                provides=IColumn)
class ActivitySessionsTableExternBookableColumn(IconColumn):
    """Activity sessions table external bookable column"""

    icon_class = 'far fa-calendar-alt'
    hint = _("Open for external booking")

    weight = 80

    def checker(self, item):
        return item.extern_bookable


@adapter_config(name='public',
                required=(IWfPlanningTarget, IAdminLayer, IActivitySessionsTable),
                provides=IColumn)
class ActivitySessionsTablePublicColumn(IconColumn):
    """Activity sessions table public column"""

    icon_class = 'fas fa-users'
    hint = _("Public session")

    weight = 90

    def checker(self, item):
        return item.public_session


@adapter_config(required=(ISession, IAdminLayer, IActivitySessionsTable),
                provides=ITableElementEditor)
class SessionTableElementEditor(TableElementEditor):
    """Session table element editor"""



#
# Copyright (c) 2015-2023 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_*** module

"""

from persistent import Persistent
from zope.schema.fieldproperty import FieldProperty

from pyams_app_msc.component.banking.interfaces import IBankingAccount
from pyams_utils.factory import factory_config

__docformat__ = 'restructuredtext'


@factory_config(IBankingAccount)
class BankingAccount(Persistent):
    """Banking account persistent class"""

    company_name = FieldProperty(IBankingAccount['company_name'])
    bank_code = FieldProperty(IBankingAccount['bank_code'])
    counter_code = FieldProperty(IBankingAccount['counter_code'])
    account_number = FieldProperty(IBankingAccount['account_number'])
    account_key = FieldProperty(IBankingAccount['account_key'])
    iban_number = FieldProperty(IBankingAccount['iban_number'])
    bic_code = FieldProperty(IBankingAccount['bic_code'])



# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

from ZODB.POSException import POSKeyError
from ZODB.utils import p64
from cornice import Service
from cornice.validators import colander_validator
from hypatia.catalog import CatalogQuery
from hypatia.interfaces import ICatalog
from hypatia.query import And, Any, Eq, Or
from pyramid.httpexceptions import HTTPBadRequest, HTTPForbidden, HTTPNotFound, HTTPOk, HTTPUnauthorized
from zope.intid.interfaces import IIntIds
from zope.lifecycleevent import ObjectModifiedEvent
from zope.schema.interfaces import ValidationError

from pyams_app_msc.feature.booking import BOOKING_STATUS, IBookingInfo
from pyams_app_msc.feature.booking.api.interfaces import MSC_BOOKING_API_ROUTE, MSC_BOOKING_SEARCH_API_ROUTE, \
    MSC_BOOKING_VALIDATION_API_ROUTE
from pyams_app_msc.feature.booking.api.schema import BookingInfoRequest, BookingSearchRequest, BookingSearchResponse, \
    BookingValidationGetterResponse, BookingValidationPostRequest, BookingValidationPostResponse, \
    BookingValidationRequest, FullBookingInfoResponse
from pyams_app_msc.feature.booking.message import get_booking_message
from pyams_app_msc.feature.messaging import IMessagingSettings
from pyams_app_msc.feature.planning.interfaces import ISession
from pyams_app_msc.feature.profile import IOperatorProfile
from pyams_app_msc.interfaces import MANAGE_BOOKING_PERMISSION, VIEW_BOOKING_PERMISSION
from pyams_app_msc.shared.theater import IMovieTheater
from pyams_catalog.query import CatalogResultSet
from pyams_content.feature.history import IHistoryContainer
from pyams_content_api.feature.json import IJSONExporter
from pyams_security.interfaces.base import USE_INTERNAL_API_PERMISSION
from pyams_security.rest import check_cors_origin, set_cors_headers
from pyams_utils.factory import get_interface_base_name
from pyams_utils.registry import get_utility
from pyams_utils.rest import STATUS, http_error, rest_responses

__docformat__ = 'restructuredtext'

from pyams_app_msc import _


booking_search_service = Service(name=MSC_BOOKING_SEARCH_API_ROUTE,
                                 pyramid_route=MSC_BOOKING_SEARCH_API_ROUTE,
                                 description="Booking search service")


@booking_search_service.options(validators=(check_cors_origin, set_cors_headers))
def booking_search_options(request):
    """Booking search service options"""
    return ''


booking_search_responses = rest_responses.copy()
booking_search_responses[HTTPOk.code] = BookingSearchResponse(description="Booking search response")


@booking_search_service.get(schema=BookingSearchRequest(),
                            validators=(check_cors_origin, colander_validator, set_cors_headers),
                            response_schemas=booking_search_responses)
def get_bookings(request):
    """Get bookings list"""
    if not request.has_permission(USE_INTERNAL_API_PERMISSION, context=request.context):
        return http_error(request, HTTPUnauthorized)
    registry = request.registry
    profile = IOperatorProfile(request)
    catalog = get_utility(ICatalog)
    intids = get_utility(IIntIds)
    principal_id = request.principal.id
    params = And(Eq(catalog['object_types'], get_interface_base_name(IMovieTheater)),
                 Or(Eq(catalog['role:msc:manager'], principal_id),
                    Eq(catalog['role:msc:operator'], principal_id),
                    Eq(catalog['role:msc:contributor'], principal_id),
                    Eq(catalog['role:msc:reader'], principal_id)))
    bookings = []
    theaters = list(CatalogResultSet(CatalogQuery(catalog).query(params)))
    if not theaters:
        translate = request.localizer.translate
        return {
            'status': STATUS.SUCCESS.value,
            'message': translate(_("You are not assigned to any theater")),
            'results': bookings
        }
    append_booking = bookings.append
    params = And(Eq(catalog['object_types'], get_interface_base_name(IBookingInfo)),
                 Any(catalog['parents'], [intids.queryId(theater) for theater in theaters]),
                 Any(catalog['booking_status'], {
                     BOOKING_STATUS.OPTION.value,
                     BOOKING_STATUS.WAITING.value
                 }))
    request_params = request.validated.get('querystring', {})
    for booking in sorted(filter(lambda x: not x.archived,
                                 CatalogResultSet(CatalogQuery(catalog).query(params))),
                          key=lambda x: ISession(x).start_date):
        exporter = registry.queryMultiAdapter((booking, request), IJSONExporter,
                                              name='msc_mobile_api')
        if exporter is not None:
            append_booking(exporter.to_json(**request_params))
    return {
        'status': STATUS.SUCCESS.value,
        'results': bookings
    }


booking_info_service = Service(name=MSC_BOOKING_API_ROUTE,
                               pyramid_route=MSC_BOOKING_API_ROUTE,
                               description="Booking info service")


@booking_info_service.options(validators=(check_cors_origin, set_cors_headers))
def booking_info_options(request):
    """Booking info service options"""
    return ''


booking_info_responses = rest_responses.copy()
booking_info_responses[HTTPOk.code] = FullBookingInfoResponse(description="Booking info response")


@booking_info_service.get(schema=BookingInfoRequest(),
                          validators=(check_cors_origin, colander_validator, set_cors_headers),
                          response_schemas=booking_info_responses)
def get_booking_info(request):
    """Booking information getter"""
    if not request.has_permission(USE_INTERNAL_API_PERMISSION, context=request.context):
        return http_error(request, HTTPUnauthorized)
    booking_id = request.matchdict['booking_id']
    if not booking_id:
        return http_error(request, HTTPBadRequest, "Missing booking ID")
    try:
        booking = request.context._p_jar.get(p64(int(booking_id, 16)))
    except ValueError:
        return http_error(request, HTTPBadRequest, "Bad session ID")
    except POSKeyError:
        return http_error(request, HTTPNotFound)
    if not IBookingInfo.providedBy(booking):
        return http_error(request, HTTPBadRequest, "Bad session ID")
    if not request.has_permission(VIEW_BOOKING_PERMISSION, context=booking):
        return {
            'status': STATUS.SUCCESS.value,
            'message': _("You are not allowed to access this booking"),
            'booking': None
        }
    request_params = request.validated.get('querystring', {})
    exporter = request.registry.queryMultiAdapter((booking, request), IJSONExporter,
                                                  name='msc_mobile_api')
    if exporter is not None:
        return {
            'status': STATUS.SUCCESS.value,
            'booking': exporter.to_json(**request_params)
        }
    return http_error(request, HTTPNotFound)


#
# Booking validation API
#

booking_validation_service = Service(name=MSC_BOOKING_VALIDATION_API_ROUTE,
                                     pyramid_route=MSC_BOOKING_VALIDATION_API_ROUTE,
                                     description="Booking validation service")


@booking_validation_service.options(validators=(check_cors_origin, set_cors_headers))
def booking_validation_options(request):
    """Booking validation service options"""
    return ''


booking_validation_getter_responses = rest_responses.copy()
booking_validation_getter_responses[HTTPOk.code] = BookingValidationGetterResponse(
        description="Booking validation response")


@booking_validation_service.get(schema=BookingValidationRequest(),
                                validators=(check_cors_origin, colander_validator, set_cors_headers),
                                response_schemas=booking_validation_getter_responses)
def get_booking_validation_info(request):
    """Booking validation information getter"""
    if not request.has_permission(USE_INTERNAL_API_PERMISSION, context=request.context):
        return http_error(request, HTTPUnauthorized)
    booking_id = request.matchdict['booking_id']
    if not booking_id:
        return http_error(request, HTTPBadRequest, "Missing booking ID")
    try:
        booking = request.context._p_jar.get(p64(int(booking_id, 16)))
    except ValueError:
        return http_error(request, HTTPBadRequest, "Bad session ID")
    except POSKeyError:
        return http_error(request, HTTPNotFound)
    if not IBookingInfo.providedBy(booking):
        return http_error(request, HTTPBadRequest, "Bad session ID")
    if not request.has_permission(MANAGE_BOOKING_PERMISSION, context=booking):
        return {
            'status': STATUS.SUCCESS.value,
            'message': _("You are not allowed to manage this booking"),
            'booking': None
        }
    request_params = request.validated.get('querystring', {})
    exporter = request.registry.queryMultiAdapter((booking, request), IJSONExporter,
                                                  name='msc_mobile_validation_api')
    if exporter is not None:
        return {
            'status': STATUS.SUCCESS.value,
            'booking': exporter.to_json(**request_params)
        }
    return http_error(request, HTTPNotFound)


booking_validation_post_responses = rest_responses.copy()
booking_validation_post_responses[HTTPOk.code] = BookingValidationPostResponse(
        description="Booking validation setter response")


@booking_validation_service.post(content_type='application/json',
                                 schema=BookingValidationPostRequest(),
                                 validators=(check_cors_origin, colander_validator, set_cors_headers),
                                 require_csrf=False,
                                 response_schemas=booking_validation_post_responses)
def post_booking_validation_info(request):
    """Booking validation request"""
    if not request.has_permission(USE_INTERNAL_API_PERMISSION, context=request.context):
        return http_error(request, HTTPUnauthorized)
    booking_id = request.matchdict['booking_id']
    if not booking_id:
        return http_error(request, HTTPBadRequest, "Missing booking ID")
    try:
        booking = request.context._p_jar.get(p64(int(booking_id, 16)))
    except ValueError:
        return http_error(request, HTTPBadRequest, "Bad session ID")
    except POSKeyError:
        return http_error(request, HTTPNotFound)
    if not IBookingInfo.providedBy(booking):
        return http_error(request, HTTPBadRequest, "Bad session ID")
    if not request.has_permission(MANAGE_BOOKING_PERMISSION, context=booking):
        return http_error(request, HTTPForbidden)
    history = IHistoryContainer(booking, None)
    params = request.validated.get('body', {})
    try:
        notify_message = None
        if params.get('notify_recipient'):
            notify_message = params.get('notify_message')
            if not notify_message:
                return http_error(request, HTTPBadRequest, "Missing field: notify_message")
        if history is not None:
            history.add_history(booking,
                                comment=params.get('notepad'),
                                message=notify_message,
                                request=request)
        booking.status = params['status']
        booking.accompanying_ratio = params['accompanying_ratio']
        booking.price = params['price']
        booking.notepad = params.get('notepad')
        if params.get('send_reminder'):
            booking.reminder_subject = params.get('reminder_subject')
            booking.reminder_message = params.get('reminder_message')
        if params.get('notify_recipient'):
            if params.get('include_quotation'):
                message = params.get('quotation_message')
                if message:
                    booking.quotation_message = message
            settings = IMessagingSettings(request.root, None)
            if settings is not None:
                mailer = settings.get_mailer()
                if mailer is not None:
                    subject = _("Booking accepted") if booking.status == BOOKING_STATUS.ACCEPTED.value \
                        else _("Booking temporarily accepted")
                    html_message = get_booking_message(subject, params, booking, request, settings)
                    if html_message is not None:
                        mailer.send(html_message)
        request.registry.notify(ObjectModifiedEvent(booking))
        return {
            'status': STATUS.SUCCESS.value
        }
    except ValidationError as exc:
        return http_error(request, HTTPBadRequest, f"Invalid value for field: {exc.field.__name__}")
    


# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

from colander import Boolean, Integer, MappingSchema, SchemaNode, SequenceSchema, String, drop

from pyams_app_msc.feature.booking.api.schema import BookingList
from pyams_utils.rest import BaseResponseSchema


__docformat__ = 'restructuredtext'


#
# Planning API schemas
#

class PlanningSessionInfo(MappingSchema):
    """Planning session info"""
    id = SchemaNode(String(),
                    description="Session ID")
    title = SchemaNode(String(),
                       description="Session title")
    start = SchemaNode(String(),
                          description="Session start date")
    end = SchemaNode(String(),
                     description="Session end date")
    theater = SchemaNode(String(),
                         description="Theater name")
    room = SchemaNode(String(),
                      description="Room name")
    capacity = SchemaNode(Integer(),
                          description="Room capacity")
    temporary = SchemaNode(Boolean(),
                           description="Temporary session flag")
    bookable = SchemaNode(Boolean(),
                          description="Bookable session flag")
    public = SchemaNode(Boolean(),
                        description="Public session flag")
    notepad = SchemaNode(String(),
                         description="Session notepad",
                         missing=drop)
    requested_seats = SchemaNode(Integer(),
                                 description="Requested seats",
                                 missing=drop)
    confirmed_seats = SchemaNode(Integer(),
                                 description="Confirmed seats",
                                 missing=drop)
    can_update = SchemaNode(Boolean(),
                            description="User update permission",
                            missing=drop)
    
    
class PlanningSessionsList(SequenceSchema):
    """Planning sessions list"""
    session = PlanningSessionInfo()
    
    
class PlanningGetterResults(BaseResponseSchema):
    """Planning getter result"""
    results = PlanningSessionsList(description="Planning sessions list",
                                   missing=drop)
    
    
class PlanningGetterResponse(MappingSchema):
    """Planning getter response schema"""
    body = PlanningGetterResults()


#
# Session API schemas
#

class SessionInfo(PlanningSessionInfo):
    """Session info"""
    bookings = BookingList()
    
    
class SessionResult(BaseResponseSchema):
    """Session result"""
    session = SessionInfo(description="Session info",
                          missing=drop)
    
    
class SessionGetterResponse(MappingSchema):
    """Session getter response schema"""
    body = SessionResult()



#
# Copyright (c) 2015-2023 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_*** module

"""

from datetime import datetime

from hypatia.catalog import CatalogQuery
from hypatia.interfaces import ICatalog
from hypatia.query import And, Any, Eq, Ge, Le
from zope.interface import Interface, implementer
from zope.schema import Bool, Choice, List, TextLine
from zope.schema.vocabulary import getVocabularyRegistry

from pyams_app_msc.feature.booking import IBookingInfo
from pyams_app_msc.feature.booking.interfaces import BOOKING_STATUS, BOOKING_STATUS_VOCABULARY
from pyams_app_msc.feature.booking.zmi.dashboard import BookingStatusTable, get_booking_element
from pyams_app_msc.feature.booking.zmi.interfaces import IBookingDashboardMenu
from pyams_app_msc.feature.profile import IUserProfile
from pyams_app_msc.interfaces import VIEW_BOOKING_PERMISSION
from pyams_app_msc.reference.structure import STRUCTURE_TYPES_VOCABULARY
from pyams_app_msc.shared.catalog import ICatalogEntry, IWfCatalogEntry
from pyams_app_msc.shared.theater import IMovieTheater
from pyams_app_msc.shared.theater.interfaces.room import ROOMS_SEATS_VOCABULARY
from pyams_app_msc.shared.theater.interfaces.session import IMovieTheaterSession
from pyams_catalog.query import CatalogResultSet
from pyams_form.field import Fields
from pyams_form.interfaces.form import IFormFields
from pyams_layer.interfaces import IPyAMSLayer
from pyams_pagelet.pagelet import pagelet_config
from pyams_security.schema import PrincipalsSetField
from pyams_sequence.interfaces import ISequentialIdInfo
from pyams_sequence.schema import InternalReferencesListField
from pyams_skin.interfaces.viewlet import IHeaderViewletManager
from pyams_table.column import GetAttrColumn
from pyams_table.interfaces import IColumn, IValues
from pyams_utils.adapter import ContextRequestViewAdapter, adapter_config
from pyams_utils.factory import get_interface_base_name
from pyams_utils.interfaces import MISSING_INFO
from pyams_utils.interfaces.data import IObjectData
from pyams_utils.list import unique_iter
from pyams_utils.registry import get_utility
from pyams_utils.schema import DatesRangeField
from pyams_utils.timezone import tztime
from pyams_utils.traversing import get_parent
from pyams_viewlet.viewlet import EmptyViewlet, viewlet_config
from pyams_zmi.interfaces import IAdminLayer
from pyams_zmi.search import SearchForm, SearchResultsView, SearchView
from pyams_zmi.table import I18nColumnMixin
from pyams_zmi.zmi.viewlet.menu import NavigationMenuItem

__docformat__ = 'restructuredtext'

from pyams_app_msc import _


@viewlet_config(name='advanced-booking-search.menu',
                context=IMovieTheater, layer=IAdminLayer,
                manager=IBookingDashboardMenu, weight=40,
                permission=VIEW_BOOKING_PERMISSION)
class BookingAdvancedSearchMenu(NavigationMenuItem):
    """Booking advanced search menu"""

    label = _("Advanced search")
    href = '#booking-advanced-search.html'


class IBookingAdvancedSearchQuery(Interface):
    """Booking advanced search query interface"""

    recipients = PrincipalsSetField(title=_("Recipients"),
                                    required=False)

    activities = InternalReferencesListField(title=_("Activities"),
                                             required=False)

    establishment = TextLine(title=_("Establishment"),
                             required=False)

    city = TextLine(title=_("City"),
                    required=False)

    structure_types = List(title=_("Structures types"),
                           value_type=Choice(vocabulary=STRUCTURE_TYPES_VOCABULARY),
                           required=False)

    status = List(title=_("Status"),
                  value_type=Choice(vocabulary=BOOKING_STATUS_VOCABULARY),
                  required=False)

    include_archives = Bool(title=_("Include archives"),
                            required=False,
                            default=False)

    session = DatesRangeField(title=_("Session date"),
                              required=False)

    created = DatesRangeField(title=_("Creation date"),
                              required=False)

    modified = DatesRangeField(title=_("Modification date"),
                               required=False)


class BookingAdvancedSearchForm(SearchForm):
    """Booking advanced search form"""

    title = _("Bookings search form")

    ajax_form_handler = 'booking-advanced-search-results.html'
    _edit_permission = VIEW_BOOKING_PERMISSION


@adapter_config(required=(Interface, IAdminLayer, BookingAdvancedSearchForm),
                provides=IFormFields)
def booking_advanced_search_form_fields(context, request, form):
    """Booking advanced search form fields"""
    return Fields(IBookingAdvancedSearchQuery)


@pagelet_config(name='booking-advanced-search.html',
                context=IMovieTheater, layer=IPyAMSLayer,
                permission=VIEW_BOOKING_PERMISSION)
class BookingAdvancedSearchView(SearchView):
    """Booking advanced search view"""

    title = _("Bookings search form")
    header_label = _("Advanced search")
    search_form = BookingAdvancedSearchForm


@implementer(IObjectData)
class BookingAdvancedSearchResultsTable(BookingStatusTable):
    """Booking advanced search form results table"""

    object_data = {
        'buttons': ['colvis', 'copy', 'csv', 'excel', 'print'],
        'ams-buttons-classname': 'btn btn-sm btn-secondary'
    }


@adapter_config(name='establishment',
                required=(IMovieTheater, IAdminLayer, BookingAdvancedSearchResultsTable),
                provides=IColumn)
@implementer(IObjectData)
class BookingStatusEstablishmentColumn(I18nColumnMixin, GetAttrColumn):
    """Booking status establishment column"""

    i18n_header = _("Establishment")
    object_data = {
        'visible': False
    }
    css_classes = {
        'td': 'text-nowrap'
    }
    weight = 26

    def get_value(self, obj):
        profile = IUserProfile(obj.booking, None)
        if profile is not None:
            return profile.establishment
        return MISSING_INFO


@adapter_config(name='structure_type',
                required=(IMovieTheater, IAdminLayer, BookingAdvancedSearchResultsTable),
                provides=IColumn)
@implementer(IObjectData)
class BookingStatusStructureTypeColumn(I18nColumnMixin, GetAttrColumn):
    """Booking status structure type column"""

    i18n_header = _("Structure type")
    object_data = {
        'visible': False
    }
    css_classes = {
        'td': 'text-nowrap'
    }
    weight = 27

    def get_value(self, obj):
        profile = IUserProfile(obj.booking, None)
        return profile.get_structure_type() if profile is not None else MISSING_INFO


@adapter_config(name='city',
                required=(IMovieTheater, IAdminLayer, BookingAdvancedSearchResultsTable),
                provides=IColumn)
@implementer(IObjectData)
class BookingStatusCityColumn(I18nColumnMixin, GetAttrColumn):
    """Booking status city column"""

    i18n_header = _("City")
    object_data = {
        'visible': False
    }
    css_classes = {
        'td': 'text-nowrap'
    }
    weight = 28

    def get_value(self, obj):
        profile = IUserProfile(obj.booking, None)
        if profile is not None:
            address = profile.establishment_address
            return address.city if address is not None else MISSING_INFO
        return MISSING_INFO


@adapter_config(required=(IMovieTheater, IPyAMSLayer, BookingAdvancedSearchResultsTable),
                provides=IValues)
class BookingAdvancedSearchResultsValues(ContextRequestViewAdapter):
    """Booking advanced search results values"""

    def get_params(self, data):
        catalog = get_utility(ICatalog)
        vocabulary = getVocabularyRegistry().get(self.context, ROOMS_SEATS_VOCABULARY)
        params = And(Eq(catalog['object_types'], get_interface_base_name(IBookingInfo)),
                     Any(catalog['planning_room'], vocabulary.by_value.keys()),
                     Eq(catalog['booking_status'], [status.value for status in BOOKING_STATUS]))
        if data.get('recipients'):
            params &= Any(catalog['booking_recipient'], data['recipients'])
        if data.get('status'):
            params &= Any(catalog['booking_status'], data['status'])
        session_after, session_before = data.get('session', (None, None))
        if session_after:
            params &= Ge(catalog['planning_start_date'],
                         tztime(datetime.fromisoformat(session_after.isoformat())))
        if session_before:
            params &= Le(catalog['planning_end_date'],
                         tztime(datetime.fromisoformat(session_before.isoformat())))
        created_after, created_before = data.get('created', (None, None))
        if created_after:
            params &= Ge(catalog['created_date'],
                         tztime(datetime.fromisoformat(created_after.isoformat())))
        if created_before:
            params &= Le(catalog['created_date'],
                         tztime(datetime.fromisoformat(created_before.isoformat())))
        modified_after, modified_before = data.get('modified', (None, None))
        if modified_after:
            params &= Ge(catalog['modified_date'],
                         tztime(datetime.fromisoformat(modified_after.isoformat())))
        if modified_before:
            params &= Le(catalog['modified_date'],
                         tztime(datetime.fromisoformat(modified_before.isoformat())))
        return params

    @property
    def values(self):
        """Booking advanced search results values getter"""

        def booking_filter(item):
            booking = item.booking
            if booking.archived and (not data.get('include_archives')):
                return False
            session = item.session
            oids = data.get('activities')
            if oids:
                if IMovieTheaterSession.providedBy(session):
                    catalog_entry = IWfCatalogEntry(session, None)
                else:
                    catalog_entry = get_parent(session, ICatalogEntry)
                if catalog_entry is None:
                    return False
                if ISequentialIdInfo(catalog_entry).hex_oid not in oids:
                    return False
            profile = IUserProfile(booking, None)
            if profile is None:
                return False
            establishment = data.get('establishment')
            if establishment and (
                    (not profile.establishment) or
                    (establishment.lower() not in profile.establishment.lower())):
                return False
            city = data.get('city')
            if city:
                if not profile.establishment_address:
                    return False
                if city.lower() not in profile.establishment_address.city.lower():
                    return False
            structure_types = data.get('structure_types')
            if structure_types and (profile.structure_type not in structure_types):
                return False
            return True

        form = BookingAdvancedSearchForm(self.context, self.request)
        form.update()
        data, _errors = form.extract_data()
        params = self.get_params(data)
        catalog = get_utility(ICatalog)
        yield from filter(booking_filter,
                          map(get_booking_element,
                              unique_iter(CatalogResultSet(CatalogQuery(catalog).query(params)))))


@pagelet_config(name='booking-advanced-search-results.html',
                context=IMovieTheater, layer=IPyAMSLayer,
                permission=VIEW_BOOKING_PERMISSION, xhr=True)
class BookingAdvancedSearchResultsView(SearchResultsView):
    """Booking advanced search results view"""

    table_label = _("Search results")
    table_class = BookingAdvancedSearchResultsTable


@viewlet_config(name='pyams.content_header',
                layer=IAdminLayer, view=BookingAdvancedSearchResultsView,
                manager=IHeaderViewletManager, weight=10)
class BookingAdvancedSearchResultsViewHeaderViewlet(EmptyViewlet):
    """Booking advanced search results view header viewlet"""

    def render(self):
        return '<h1 class="mt-3"></h1>'



# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

from pyams_app_msc.feature.planning.interfaces import ISession
from pyams_app_msc.shared.catalog.interfaces import CATALOG_ENTRY_CONTENT_TYPE
from pyams_sequence.schema import InternalReferenceField

__docformat__ = 'restructuredtext'

from pyams_app_msc import _


class IMovieTheaterSession(ISession):
    """Movie theater session interface"""

    activity = InternalReferenceField(title=_("Activity"),
                                      description=_("You can select an existing activity from your catalog, or keep "
                                                    "this input empty if you just want to mark a period as not being "
                                                    "available for booking"),
                                      content_type=CATALOG_ENTRY_CONTENT_TYPE,
                                      required=False)
